from typing import Union, List, Annotated

import numpy as np
import pandas as pd
import math
import random


class FbmSimulator:
    """
        Source paper: todo:add source paper
        Corresponding Matlab code: todo:add link to matlab code
        Main idea: we use Lamperti transform to transfer FBM (self-similar process) to a stationary process, and
                   simulate the stationary process using circulant embedding approach (Wood, A.T.A., Chan, G., 1994.
                   Simulation of stationary Gaussian processes in [0, 1]^d. Journal of computational and graphical
                   statistics 3, 409–432). Then a subsequence of the simulated stationary process is convert back to
                   the FBM series. Because we need to use its subsequence, the stationary process needs to have a longer
                   length (lamperti_series_len_multiplier*series_len) than the FBM series.
    """

    def __init__(self,
                 series_len: int,
                 hurst_parameter: float,
                 lamperti_series_len_multiplier: int = 5,
                 tmax: float = 1,
                 const: float = 1,
                 seed: int = None):
        """
            series_len: length of the fbm series to simulate
            hurst_index: hurst parameter of fbm
            lamperti_series_len_multiplier: bigger value (usually <=10) provides more accuracy; default value is 5
            tmax: the time runs in [0,tmax]; the default value is 1
            const: standard derivation at initial point; default value is 1

        """
        self.series_len = series_len
        self.hurst_parameter = hurst_parameter
        if not (0 < self.hurst_parameter < 1):
            raise ValueError(f'Hurst index must be in range (0, 1).')
        self.tmax = tmax
        if not self.tmax > 0:
            raise ValueError(f'tmax must be positive.')
        self.const = const
        if not self.const > 0:
            raise ValueError(f'const must be positive.')
        self.seed = seed
        self.lamperti_series_len_multiplier = lamperti_series_len_multiplier
        self.lamperti_series_len = self.lamperti_series_len_multiplier * self.series_len

    @property
    def _lamperti_subseq_index(self):
        seires_step = self.tmax / self.series_len
        series_t = np.arange(start=seires_step, stop=self.tmax + seires_step, step=seires_step)
        # transfer time to (1,e)
        series_exp_t = (np.exp(1) - 1)/self.tmax * series_t + 1
        # log of time in (1,e) is in (0,1)
        log_series_exp_t = np.log(series_exp_t)
        max_log_series_exp_t = np.max(log_series_exp_t)
        lamperti_subseq_index = np.rint(log_series_exp_t * self.lamperti_series_len / max_log_series_exp_t) - 1
        return lamperti_subseq_index.astype(int)

    def _line_circulant(self, m):
        """
            First line of the circulant matrix C built with covariances of the stationary process.
        """


from typing import Union, List

import numpy as np
import pandas as pd
import math


class IrHurstEstimator:
    """
        Source paper: Bardet, Jean-Marc & Surgailis, Donatas, 2013.
                      "Nonparametric estimation of the local Hurst function of multifractional Gaussian processes,"
                      Stochastic Processes and their Applications, Elsevier, vol. 123(3), pages 1004-1045.
        Original Matlab code source: http://samm.univ-paris1.fr/Sofwares-Logiciels
                                     Software for estimating the Hurst function H of a Multifractional Brownian Motion:
                                     Quadratic Variation estimator and IR estimator
    """

    def __init__(self, mbm_series: Union[List, np.ndarray, pd.Series], alpha: float):
        """
            mbm_series: multifractional Brownian motion series
            alpha: decide how many observations on mbm_series is used to estimate a point of the holder exponent;
                   small alpha means more observations are used for a single point and therefore the variance is small.
        """
        self.mbm_series = np.array(mbm_series)
        self.alpha = alpha
        if not (0 < self.alpha < 1):
            raise ValueError(f'Alpha must be in range (0, 1).')
        self.n = len(self.mbm_series)

    @staticmethod
    def _ismember(a, b):
        bind = {}
        for i, elt in enumerate(b):
            if elt not in bind:
                bind[elt] = i
        return [bind.get(itm, None) for itm in a]  # None can be replaced by any other "not in b" value

    def _neighborhood(self, t: int, step: int = 1):
        """
            The function $\mathcal{V}_{n,\alpha}(t)$ in Sec 2.1 of the source paper:
                      Bardet, Jean-Marc & Surgailis, Donatas, 2013.
                      "Nonparametric estimation of the local Hurst function of multifractional Gaussian processes,"
                      Stochastic Processes and their Applications, Elsevier, vol. 123(3), pages 1004-1045.
        """
        n_to_1_minus_alpha = self.n ** (1 - self.alpha)
        k_min = max(0.0, math.ceil(t - n_to_1_minus_alpha))  # the pdep in matlab code
        k_max = min(float(self.n - 3 - 2), math.floor(t + n_to_1_minus_alpha))  # the pfin in matlab code
        return np.arange(start=k_min, stop=k_max + 1, step=step, dtype=int)

    @staticmethod
    def _hest_ir(R, p: int):
        """
            Equivalent to matlab function HestIR(R,p) in the source code:
                        http://samm.univ-paris1.fr/Sofwares-Logiciels
                        Software for estimating the Hurst function H of a Multifractional Brownian Motion:
                        Quadratic Variation estimator and IR estimator
            This is $\Lambda_2^{-1}(H)$ (inverse function of eq. 2.5) in the paper:
                        Bardet, Jean-Marc & Surgailis, Donatas, 2013.
                        "Nonparametric estimation of the local Hurst function of multifractional Gaussian processes,"
                        Stochastic Processes and their Applications, Elsevier, vol. 123(3), pages 1004-1045.
        """
        start = 0.0001
        stop = 0.9999
        step = 0.0001
        H = np.arange(start, stop + step, step)
        if p == 1:
            r = 2 ** (2 * H - 1) - 1
        elif p == 2:
            r = (-3 ** (2 * H) + 2 ** (2 * H + 2) - 7) / (8 - 2 ** (2 * H + 1))
        else:
            raise ValueError('p is the order of the IR statistic; only takes 1 or 2.')
        lam = 1 / np.pi * np.arccos(-r) + np.sqrt(1 - r ** 2) / (np.pi * (1 - r)) * np.log(2 / (r + 1))
        return sum(R > i for i in lam) / 10000

    @property
    def holder_exponents(self):
        mgn_series = self.mbm_series[1:] - self.mbm_series[:-1]
        ir_estimates = []
        for t in range(self.n):
            k = self._neighborhood(t=t)
            R2 = np.mean(np.abs(mgn_series[3 + k] - mgn_series[1 + k]) / (
                    np.abs(mgn_series[3 + k] - mgn_series[2 + k]) + np.abs(mgn_series[2 + k] - mgn_series[1 + k])))
            ir_estimates.append(self._hest_ir(R=R2, p=2))
        return np.array(ir_estimates)


class QvHurstEstimator(IrHurstEstimator):
    """
        Source paper: Bardet, Jean-Marc & Surgailis, Donatas, 2013.
                      "Nonparametric estimation of the local Hurst function of multifractional Gaussian processes,"
                      Stochastic Processes and their Applications, Elsevier, vol. 123(3), pages 1004-1045.
        Original Matlab code source: http://samm.univ-paris1.fr/Sofwares-Logiciels
                                     Software for estimating the Hurst function H of a Multifractional Brownian Motion:
                                     Quadratic Variation estimator and IR estimator
    """

    def __init__(self, mbm_series: Union[List, np.ndarray, pd.Series], alpha: float, p: int = 5):
        """
            mbm_series: multifractional Brownian motion series
            alpha: decide how many observations on mbm_series is used to estimate a point of the holder exponent;
                   small alpha means more observations are used for a single point and therefore the variance is small.
            p: the number of dilatations of QV estimates and it is 5 according to the paper.
        """
        super().__init__(mbm_series=mbm_series, alpha=alpha)
        self.p = p

    def _neighborhood(self, t: int, step: int = 1):
        """
            Modified (for QV) function $\mathcal{V}_{n,\alpha}(t)$ in Sec 2.1 of the source paper:
                      Bardet, Jean-Marc & Surgailis, Donatas, 2013.
                      "Nonparametric estimation of the local Hurst function of multifractional Gaussian processes,"
                      Stochastic Processes and their Applications, Elsevier, vol. 123(3), pages 1004-1045.
        """
        n_to_1_minus_alpha = self.n ** (1 - self.alpha)
        k_min = max(0.0, math.ceil(t - n_to_1_minus_alpha))  # the pdep in matlab code
        k_max = min(float(self.n - 11), math.floor(t + n_to_1_minus_alpha))  # the pfin in matlab code
        return np.arange(start=k_min, stop=k_max + 1, step=step, dtype=int)

    def _get_S(self, t):
        """
            Eq (2.7): i=1,...self.p
        """
        k = self._neighborhood(t=t, step=1)  # the matlab code uses a neighborhood of every i steps, here we use step=1
        S_lst = []
        for i in range(1, self.p + 1):
            S_lst.append(np.mean((self.mbm_series[2 * i + k] - 2 * self.mbm_series[i + k] + self.mbm_series[k]) ** 2))
        return np.array(S_lst)

    @staticmethod
    def _get_A(p: int = 5):
        log_lst = np.log(range(1, p + 1))
        mean_log_lst = np.mean(log_lst)
        return log_lst - mean_log_lst

    @property
    def holder_exponents(self):
        A_lst = self._get_A()
        AAT = np.dot(A_lst, A_lst.T)
        A_over_AAT_over2 = A_lst / 2 / AAT
        qv_estimates = []
        for t in range(self.n):
            log_S_lst = np.log(self._get_S(t=t))
            qv_estimates.append(np.dot(A_over_AAT_over2, log_S_lst))
        return np.array(qv_estimates)



import functools
import math
from abc import ABC, abstractmethod
from typing import Union, List

import numpy as np
from numpy import linalg
from scipy.linalg import sqrtm


class CriticalSurfaceBrownianMotion(ABC):
    """
        alpha: significance level (look at quantiles of order alpha/2 and 1 − alpha/2)
    """

    def __init__(self,
                 N: int,
                 alpha: float = 0.05,
                 k: int = 1,
                 chi2_trials: int = 100000,
                 is_increment_series: bool = False):
        self.N = N
        self.alpha = alpha
        self.k = k
        np.random.seed(6)
        self.chi2 = np.random.chisquare(df=1, size=(self.N, chi2_trials))
        np.random.seed()
        self.is_increment_series = is_increment_series

    @functools.cached_property
    def matrix_A_k(self):
        if self.k == 0:
            A_k = np.diag(np.ones(4) / self.N)
        else:
            A_k = np.zeros((self.N, self.N))
            for i in range(self.N):
                for j in range(self.N):
                    if abs(i - j) == self.k:
                        A_k[i, j] = 0.5 / (self.N - self.k)
        return A_k

    def _add_on_sig2_matrix(self, add_on_sig2: float):
        add_on_sig2_matrix = np.zeros((self.N, self.N))
        for i in range(self.N):
            for j in range(i, self.N):
                k = j - i
                if k == 0:
                    add_on_sig2_matrix[i, j] = 2 * add_on_sig2
                    add_on_sig2_matrix[j, i] = 2 * add_on_sig2
                if k == 1:
                    add_on_sig2_matrix[i, j] = - add_on_sig2
                    add_on_sig2_matrix[j, i] = - add_on_sig2
        return add_on_sig2_matrix

    @abstractmethod
    def _autocovariance_matrix_increment(self, sig2: float, H: Union[List[float], np.ndarray, float],
                                         add_on_sig2: float):
        pass

    @abstractmethod
    def _autocovariance_matrix(self, sig2: float, H: Union[List[float], np.ndarray, float], add_on_sig2: float):
        pass

    def _eigenvalues(self, sig2: float, H: Union[List[float], np.ndarray, float], add_on_sig2: float):
        if self.is_increment_series:
            Sigma: np.ndarray = self._autocovariance_matrix_increment(sig2=sig2, H=H, add_on_sig2=add_on_sig2)
        else:
            Sigma: np.ndarray = self._autocovariance_matrix(sig2=sig2, H=H, add_on_sig2=add_on_sig2)
        square_root_Sigma = sqrtm(Sigma)
        square_root_Sigma = np.real(square_root_Sigma)
        A_k = self.matrix_A_k
        mat = square_root_Sigma.dot(A_k)
        mat = mat.dot(square_root_Sigma)
        eigenvalues, _ = linalg.eig(mat)
        eigenvalues = np.real(eigenvalues)
        return eigenvalues

    def _generalized_chi2(self, sig2, H: Union[List[float], np.ndarray, float], add_on_sig2: float):
        eigenvalues = self._eigenvalues(sig2=sig2, H=H, add_on_sig2=add_on_sig2)
        stat = np.array(eigenvalues.dot(self.chi2))
        return stat

    def quantile(self, sig2, H: Union[List[float], np.ndarray, float], add_on_sig2: float = 0):
        dist = self._generalized_chi2(sig2=sig2, H=H, add_on_sig2=add_on_sig2)
        quantile = [0, 0]
        quantile[0] = np.quantile(a=dist, q=self.alpha / 2)
        quantile[1] = np.quantile(a=dist, q=1 - self.alpha / 2)
        return quantile


class CriticalSurfaceFBM(CriticalSurfaceBrownianMotion):
    """
        Source paper: Michał Balcerek, Krzysztof Burnecki. (2020)
                      Testing of fractional Brownian motion in a noisy environment.
                      Chaos, Solitons & Fractals, Volume 140, 110097.
                      https://doi.org/10.1016/j.chaos.2020.110097
    """

    def _autocovariance_matrix(self, sig2: float, H: float, add_on_sig2: float):
        # raise ValueError("No non increment version exists, use is_increment_series=True for FBM testing.")
        if not isinstance(H, (float, int)) or H > 1 or H < 0:
            raise ValueError(f'H is {H}, but it needs to be a float in [0,1].')
        Sigma = np.zeros((self.N, self.N))
        twoH = 2 * H
        for i in range(self.N):
            for j in range(i, self.N):
                t = (i + 1) / self.N
                s = (j + 1) / self.N
                sigma = 0.5 * (t ** twoH + s ** twoH - abs(t - s) ** twoH)  # todo: /self.N
                Sigma[i, j] = sigma
                if i != j:
                    Sigma[j, i] = sigma
        return sig2 * Sigma + self._add_on_sig2_matrix(add_on_sig2=add_on_sig2)

    def _autocovariance_matrix_increment(self, sig2: float, add_on_sig2: float, H: float):
        if not isinstance(H, (float, int)) or H > 1 or H < 0:
            raise ValueError(f'H is {H}, but it needs to be a float in [0,1].')
        Sigma = np.zeros((self.N, self.N))
        twoH = 2 * H
        for i in range(self.N):
            for j in range(i, self.N):
                k = abs(i - j)
                t_k_plus_one = (k + 1) / self.N
                t_k_minus_one = (k - 1) / self.N
                t_k = k / self.N
                sigma = 0.5 * (t_k_plus_one ** twoH + abs(t_k_minus_one) ** twoH - 2 * t_k ** twoH)  # todo: /self.N
                Sigma[i, j] = sigma
                if i != j:
                    Sigma[j, i] = sigma
        return sig2 * Sigma + self._add_on_sig2_matrix(add_on_sig2=add_on_sig2)


class CriticalSurfaceMFBM(CriticalSurfaceBrownianMotion):
    """
        Source paper: Balcerek, Michał, and Krzysztof Burnecki. (2020) Testing of Multifractional Brownian Motion.
                          Entropy 22, no. 12: 1403.
                          https://doi.org/10.3390/e22121403
    """

    @functools.lru_cache(maxsize=128)
    def _D(self, x: float, y: float):
        gam_xy = math.gamma(x + y + 1)
        gam_x = math.gamma(2 * x + 1)
        gam_y = math.gamma(2 * y + 1)
        sin_xy = math.sin(math.pi * (x + y) / 2)
        sin_x = math.sin(math.pi * x)
        sin_y = math.sin(math.pi * y)
        return math.sqrt(gam_x * gam_y * sin_x * sin_y) / 2 / gam_xy / sin_xy

    def _autocovariance_matrix(self, sig2: float, H: Union[List, np.ndarray], add_on_sig2: float, delta_i: int = 0,
                               delta_j: int = 0):
        Sigma = np.zeros((self.N, self.N))
        for i in range(self.N):
            for j in range(i, self.N):
                ht_idx = min(self.N - 1, i + delta_i)
                hs_idx = min(self.N - 1, j + delta_j)
                ht = H[ht_idx]
                hs = H[hs_idx]
                hths = ht + hs
                D = self._D(x=ht, y=hs)
                # sigma = D * ((i + delta_t + 1) ** h + (j + delta_s + 1) ** h - abs(i + delta_t - j - delta_s) ** h)
                t = (i + delta_i + 1) / self.N
                s = (j + delta_j + 1) / self.N
                sigma = D * (t ** hths + s ** hths - abs(t - s) ** hths)  # todo: /self.N
                Sigma[i, j] = sigma
                if i != j:
                    Sigma[j, i] = sigma
        return sig2 * Sigma + self._add_on_sig2_matrix(add_on_sig2=add_on_sig2)

    def _autocovariance_matrix_increment(self, sig2: float, H: Union[List, np.ndarray], add_on_sig2: float):
        if isinstance(H, (float, int)):
            raise ValueError('H needs to be a list for MBM.')
        cov_t1_s1 = self._autocovariance_matrix(delta_i=1, delta_j=1, H=H, sig2=1, add_on_sig2=0)
        cov_t1_s0 = self._autocovariance_matrix(delta_i=1, delta_j=0, H=H, sig2=1, add_on_sig2=0)
        cov_t0_s1 = self._autocovariance_matrix(delta_i=0, delta_j=1, H=H, sig2=1, add_on_sig2=0)
        cov_t0_s0 = self._autocovariance_matrix(delta_i=0, delta_j=0, H=H, sig2=1, add_on_sig2=0)
        cov_inc = cov_t1_s1 - cov_t1_s0 - cov_t0_s1 + cov_t0_s0
        return sig2 * cov_inc + self._add_on_sig2_matrix(add_on_sig2=add_on_sig2)



from typing import Union, List

import numpy as np
import pandas as pd

from fractal_analysis.tester.critical_surface import CriticalSurfaceBrownianMotion, CriticalSurfaceFBM, \
    CriticalSurfaceMFBM
from fractal_analysis.tester.estimate_sigma import EstimateSigma


class SeriesTester:
    def __init__(self,
                 critical_surface: CriticalSurfaceBrownianMotion,
                 is_cache_stat=False,
                 is_cache_quantile=False):
        """
        is_cache_stat: allow cache variable stat, set to True if x is unchanged
        is_cache_quantile: allow cache variable quantile, set to True if sig2 and h are unchanged
        """
        self.critical_surface = critical_surface
        self.is_cache_stat = is_cache_stat
        self.is_cache_quantile = is_cache_quantile
        self._old_series = None
        self._old_h = None
        self._old_sig2 = None
        self._stat = None
        self._quantile = None

    def test(self, x: Union[List, np.ndarray, pd.Series], h: Union[float, List, np.ndarray, pd.Series],
             add_on_sig2: float = 0, sig2: float = None):
        """
            x: series to test
            h: holder exponent, same length as x if test MBM, float in (0,1) if test FBM
            sig2: sigma square, if None, use auto-estimated sigma
            add_on_sig2: add on noise
        """
        if self.critical_surface.is_increment_series:
            series = np.diff(x, prepend=0)
        else:
            series = np.array(x)
        if isinstance(self.critical_surface, CriticalSurfaceFBM):
            if not isinstance(h, (float, int)):
                raise ValueError("h must be float for FBM tester.")
            else:
                h_series = np.ones(len(series)) * h
        elif isinstance(self.critical_surface, CriticalSurfaceMFBM):
            if isinstance(h, float):
                raise ValueError("h must be array-like for MBM tester.")
            elif isinstance(h, (List, np.ndarray, pd.Series)):
                h_series = np.array(h)
                if len(h_series) != len(series):
                    raise ValueError('h and x should have the same length for MBM tester.')
            else:
                raise ValueError("h must be List or np.ndarray or pd.Series for MBM tester.")
        else:
            raise ValueError("Critical surface type is not found.")
        if self.is_cache_stat and np.array_equal(series, self._old_series):
            stat = self._stat
        else:
            stat = np.dot(series.T.dot(self.critical_surface.matrix_A_k), series)
            if self.is_cache_stat:
                self._old_series = series
                self._stat = stat
        if sig2 is None:
            sig2 = EstimateSigma(series=series, h_series=h_series).theta_hat_square
        if self.is_cache_quantile and np.array_equal(sig2, self._old_sig2) and np.array_equal(h, self._old_h):
            quantile = self._quantile
        else:
            if isinstance(h, pd.Series):
                h = np.array(h)
            quantile = self.critical_surface.quantile(sig2=sig2, H=h, add_on_sig2=add_on_sig2)
            if self.is_cache_quantile:
                self._old_h = h
                self._old_sig2 = sig2
                self._quantile = quantile
        low = quantile[0]
        up = quantile[1]
        if low <= stat <= up:
            ret = True
        else:
            ret = False
        return ret, sig2


class MBMSeriesTester(SeriesTester):
    """
        Source paper: Balcerek, Michał, and Krzysztof Burnecki. (2020) Testing of Multifractional Brownian Motion.
                      Entropy 22, no. 12: 1403.
                      https://doi.org/10.3390/e22121403
    """

    def __init__(self, critical_surface: CriticalSurfaceMFBM, is_cache_stat=False, is_cache_quantile=False):
        super().__init__(critical_surface=critical_surface,
                         is_cache_stat=is_cache_stat,
                         is_cache_quantile=is_cache_quantile)


class FBMSeriesTester(SeriesTester):
    """
        Source paper: Michał Balcerek, Krzysztof Burnecki. (2020)
                      Testing of fractional Brownian motion in a noisy environment.
                      Chaos, Solitons & Fractals, Volume 140, 110097.
                      https://doi.org/10.1016/j.chaos.2020.110097
        Note: this class implements Algorithm 1 of the paper.
              can use an auto-estimated sigma.
    """

    def __init__(self, critical_surface: CriticalSurfaceFBM, is_cache_stat=False, is_cache_quantile=False):
        super().__init__(critical_surface=critical_surface,
                         is_cache_stat=is_cache_stat,
                         is_cache_quantile=is_cache_quantile)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fractal_analysis',
 'fractal_analysis.estimator',
 'fractal_analysis.simulator',
 'fractal_analysis.tester']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.4.0,<2.0.0', 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'fractal-analysis',
    'version': '0.1.13',
    'description': '',
    'long_description': '# Fractal Analysis\nFractal and multifractal methods, including \n\n- fractional Brownian motion (FBM) tester\n- multifractional Brownian motion (MBM) tester\n- IR hurst exponents estimator of multifractional Brownian motion (MBM)\n- QV hurst exponents estimator of multifractional Brownian motion (MBM)\n\n## FBM / MBM tester\nTest if a series is FBM (MBM) given the hurst parameter (hurst exponents series).\nThe implementation is based on the following papers:\n\n>Michał Balcerek, Krzysztof Burnecki. (2020)  \nTesting of fractional Brownian motion in a noisy environment.  \nChaos, Solitons & Fractals, Volume 140, 110097.  \nhttps://doi.org/10.1016/j.chaos.2020.110097\n\n>Balcerek, Michał, and Krzysztof Burnecki. (2020)  \nTesting of Multifractional Brownian Motion. Entropy 22, no. 12: 1403.  \nhttps://doi.org/10.3390/e22121403 \n\nWe added the following improvements to the FBM and/or MBM tester:\n- option for automatically estimating sigma \n  - based on Theorem 2.3 of the following paper:\n    >Ayache A., Peng Q. (2012)  \n     Stochastic Volatility and Multifractional Brownian Motion.  \n     In: Zili M., Filatova D. (eds) Stochastic Differential Equations and Processes. \n     Springer Proceedings in Mathematics, vol 7. Springer, Berlin, Heidelberg.  \n     https://doi.org/10.1007/978-3-642-22368-6_6\n  - a detailed introduction can be found in the section 5 of the following paper:\n    > todo: add paper name\n- option for testing if the series itself is a FBM (MBM)\n- option for testing if the increment of the series is the increment of a FBM (MBM)\n- option for testing if the series is a FBM (MBM) with an add-on noise\n- option for testing if the increment of the series is the increment of a FBM (MBM) with an add-on noise\n\n## IR / QV hurst estimator of MBM\nEstimate the hurst parameter (hurst exponent series) of a MBM.\nThe implementation is based on the following paper:\n\n>Bardet, Jean-Marc & Surgailis, Donatas, 2013.  \nNonparametric estimation of the local Hurst function of multifractional Gaussian processes.  \nStochastic Processes and their Applications, Elsevier, vol. 123(3), pages 1004-1045.\n\n\nBardet, the author in the above paper, provides a Matlab code that can be found at: \n>http://samm.univ-paris1.fr/Sofwares-Logiciels  \nSoftware for estimating the Hurst function H of a Multifractional Brownian Motion:\n Quadratic Variation estimator and IR estimator\n\n\n## To install\nTo get started, simply do:\n```\npip install fractal-analysis\n```\nor check out the code from out GitHub repository.\n\nYou can now use the series tester module in Python with:\n```\nfrom fractal_analysis import tester\n```\nor use the hurst estimator with\n```\nfrom fractal_analysis import estimator\n```\n\n## Examples\n### FBM / MBM tester\nImport:\n```\nfrom fractal_analysis.tester.series_tester import MBMSeriesTester, FBMSeriesTester\nfrom fractal_analysis.tester.critical_surface import CriticalSurfaceFBM, CriticalSurfaceMFBM\n```\nTo test if a series ```series``` is FBM, one needs to use ```CriticalSurfaceFBM``` with length of the series ```N```,\nthe significance level ```alpha``` (look at quantiles of order ```alpha/2``` and ```1 − alpha/2```), and  choose to test\non the series itself or its increment series using ```is_increment_series``` (default is ```False```, meaning to test on\nthe series itself),\n```\nfbm_tester = FBMSeriesTester(critical_surface=CriticalSurfaceFBM(N=N, alpha=0.05, is_increment_series=False))\n```\n\nTo test if the series is FBM with hurst parameter 0.3 and use auto estimated sigma square (set ```sig2=None```):\n\n```\nis_fbm, sig2 = fbm_tester.test(h=0.3, x=series, sig2=None, add_on_sig2=0)\n```\nIf the output contains, for example:\n> Bad auto sigma square calculated with error 6.239236333681868. Suggest to give sigma square and rerun.\n\nThe auto sigma square estimated is not accurate. One may want to manually choose a sigma square and rerun. For example:\n```\nis_fbm, sig2 = fbm_tester.test(h=0.3, x=series, sig2=1, add_on_sig2=0)\n```\nIf one wants to test with an add-no noise, change the value of ```add_on_sig2```.\n\n\n\n\nTo test if the series is MBM, one needs to use ```CriticalSurfaceMFBM``` with length of the series ```N```\nand the significance level ```alpha``` (look at quantiles of order ```alpha/2``` and ```1 − alpha/2```) \n```\nmbm_tester = MBMSeriesTester(critical_surface=CriticalSurfaceMFBM(N=N, alpha=0.05, is_increment_series=False))\n```\nTo test if the series is MBM with a given holder exponent series ```h_mbm_series``` and use auto estimated sigma square:\n```\nis_mbm, sig2 = mbm_tester.test(h=h_mbm_series, x=series, sig2=None, add_on_sig2=0)\n```\nBe aware that ```MBMSeriesTester``` requires ```len(h_mbm_series)==len(series)```.\n\n#### Use of cache\nUse caching to speed up the testing process. If the series ```x``` for testing is unchanged and multiple ```h``` \nand/or ```sig2``` are used, one may want to set \n```is_cache_stat=True``` to allow cache variable ```stat```. If ```h``` and ```sig2``` are unchanged and multiple ```x```\nare used, one may want to set ```is_cache_quantile=True``` to allow cache variable ```quantile```. For example:\n```\nmbm_tester = MBMSeriesTester(critical_surface=CriticalSurfaceMFBM(N=N, alpha=0.05), is_cache_stat=True, is_cache_quantile=False)\n```\n\n### IR / QV hurst estimator of MBM\nImport:\n```\nfrom fractal_analysis.estimator.hurst_estimator import IrHurstEstimator, QvHurstEstimator\nimport numpy as np\nimport math\n```\nGenerate a standard brownian motion\n```\nN = 100\nseries = np.random.randn(N) * 0.5 * math.sqrt(1 / N)\nseries = np.cumsum(series)\n```\nTo estimate the hurst exponents series of the above series with ```alpha=0.2``` using IR estimator,\n```\nestimator = IrHurstEstimator(mbm_series=series, alpha=0.2)\nprint(estimator.holder_exponents)\n```\nTo estimate the hurst exponents series of the above series with ```alpha=0.2``` using QV estimator,\n```\nestimator = QvHurstEstimator(mbm_series=series, alpha=0.2)\nprint(estimator.holder_exponents)\n```\nHere the value of ```alpha``` decides how many observations on the ```mbm_series``` is used to estimate a point of the\nholder exponent; small alpha means more observations are used for a single point and therefore the variance is small.\n   \n',
    'author': 'yujiading',
    'author_email': 'yujia.ding@cgu.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yujiading/fractals',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import cmath
import math
from typing import Union, List

import numpy as np
from scipy import integrate


class EstimateSigma:
    """
        Source paper: Ayache A., Peng Q. (2012) Stochastic Volatility and Multifractional Brownian Motion.
                      In: Zili M., Filatova D. (eds) Stochastic Differential Equations and Processes.
                      Springer Proceedings in Mathematics, vol 7. Springer, Berlin, Heidelberg.
                      https://doi.org/10.1007/978-3-642-22368-6_6
        Note: this class implements the estimator of sigma in Theorem 2.3 of the paper.
    """

    def __init__(self,
                 series: Union[List, np.ndarray],
                 h_series: Union[List, np.ndarray]):
        self.Y = [item ** 2 for item in series]  # Y is series square, p4 eq10
        self.n = len(series) - 1  # N=n  p5 line1
        self.h_series = h_series
        if len(self.h_series) != len(series):
            raise ValueError('h series and series should have the same length.')

    def C_integrant(self, eta: float, s: int):
        exp_ = cmath.exp(complex(imag=eta)) - 1
        exp_norm_sqr = exp_.real ** 2 + exp_.imag ** 2
        integrant = exp_norm_sqr ** 2 / abs(eta) ** (2 * self.h_series[s] + 3)
        return integrant

    def C(self, s: int, tol=1e-3):  # p7 eq22
        limit = 50
        max_ = 5
        text = None
        for i in range(max_):
            ret = integrate.quad(self.C_integrant, -np.inf, 0, args=(s,), limit=limit, full_output=1)
            ret = ret + integrate.quad(self.C_integrant, 0, np.inf, args=(s,), limit=limit, full_output=1)
            if ret[1] <= tol:
                return ret[0], text
            limit += 50
        text = f'Bad auto sigma square calculated with error {ret[1]}. Suggest to give sigma square and rerun.'
        return ret[0], text

    @property
    def V_bar(self):
        """
        p6 thm2.1 eq20
        h==1
        """
        text = None
        sum_ = 0
        for i in range(self.n - 1):
            C, text = self.C(s=i)
            sum_ += (self.Y[i + 1] - self.Y[i]) ** 2 / C / self.n ** (
                    -2 * self.h_series[i])
        if text:
            print(text)
        return sum_ / self.n

    @property
    def theta_hat_square(self):
        return self.V_bar * self.n / 2 / math.sqrt(math.log(self.n)) / sum(self.Y)
